import torch
import torch.nn.functional as F
import torchaudio
from transformers import AutoConfig, AutoModel, Wav2Vec2FeatureExtractor
from tqdm import tqdm
import numpy as np
import argparse
import torchmetrics
import re
from modules.args_utils import ArgsUtils
from modules.file_utils import FileUtils
from sklearn.metrics import ConfusionMatrixDisplay
import matplotlib.pyplot as plt

parser = argparse.ArgumentParser(add_help=False)

parser.add_argument('-path_validation', default='./inference_and_test_data/emo_relabeled_data_raw_improv_team_2021_4sec_v1', type=str)
parser.add_argument('-classes', default=[
    'happiness',
    'anger',
    'sadness',
    'other'
], nargs='*')
parser.add_argument('-gender', default='', type=str)
parser.add_argument('-input_size', default=64_000, type=int)  # used in inference

args, args_other = parser.parse_known_args()
args = ArgsUtils.add_other_args_v2(args, args_other)

args.device = 'cuda'
if not torch.cuda.is_available():
    args.device = 'cpu'


def speech_file_to_array_fn(path, sampling_rate):
    speech_array, _sampling_rate = torchaudio.load(path)
    resampler = torchaudio.transforms.Resample(_sampling_rate)
    speech = resampler(speech_array).squeeze().numpy()
    return speech


def predict(path, sampling_rate):
    speech = speech_file_to_array_fn(path, sampling_rate)
    inputs = feature_extractor(speech, sampling_rate=sampling_rate, return_tensors="pt", padding=True)
    inputs = {key: inputs[key].to(device) for key in inputs}

    with torch.no_grad():
        logits = model(**inputs).logits

    scores = F.softmax(logits, dim=1).detach().cpu().numpy()[0]
    predicted_class = np.argmax(scores)

    return predicted_class

def reduce_class_count(original_class):
    #'happiness 0',
    #'anger 1',
    #'sadness 2',
    #'other 3'
    reduced_class = original_class
    if original_class == 0:
        reduced_class = 1
    elif original_class == 1:
        reduced_class = 1
    elif original_class == 2:
        reduced_class = 0
    elif original_class == 3:
        reduced_class = 2
    elif original_class == 4:
        reduced_class = 0
    elif original_class == 5:
        reduced_class = 3
    elif original_class == 6:
        reduced_class = 2

    return reduced_class


config = AutoConfig.from_pretrained('Aniemore/wav2vec2-xlsr-53-russian-emotion-recognition', trust_remote_code=True)
model = AutoModel.from_pretrained("Aniemore/wav2vec2-xlsr-53-russian-emotion-recognition", trust_remote_code=True)

feature_extractor = Wav2Vec2FeatureExtractor.from_pretrained("Aniemore/wav2vec2-xlsr-53-russian-emotion-recognition")

device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
model.to(device)


eval_f1 = torchmetrics.F1Score()
eval_accuracy = torchmetrics.Accuracy()
eval_confusion_meter = torchmetrics.ConfusionMatrix(num_classes=len(args.classes), normalize="true")

samples = FileUtils.listSubFiles(args.path_validation)
true_labels = []
embeddings = []
with torch.no_grad():
    for idx, name in enumerate(tqdm(samples)):
        if idx < 2000:
            wav_name = name.split('/')[-1]
            label = re.search("happiness|sadness|anger|neutral|other", wav_name)[0]
            if label == 'neutral':
                label = 'other'

            result = predict(name, 16_000)

            reduced_class = torch.as_tensor(reduce_class_count(result))
            label = torch.as_tensor([args.classes.index(label)])

            y_prim_idx_each_long = F.one_hot(torch.LongTensor((reduced_class,)), len(args.classes)).type(torch.FloatTensor)

            eval_accuracy.update(y_prim_idx_each_long, label)
            eval_f1.update(y_prim_idx_each_long, label)
            eval_confusion_meter.update(y_prim_idx_each_long, label)

            print(f'predicted: {reduced_class}; real: {label}')


    acc =  eval_accuracy.compute().item()
    f1 = eval_f1.compute().item()
    print(f'f1: {f1}')
    print(f'accuracy: {acc}')

    conf_matrix = ConfusionMatrixDisplay(confusion_matrix=eval_confusion_meter.compute().numpy(),
                                         display_labels=args.classes)
    conf_matrix.plot(cmap=plt.cm.Blues)
    plt.savefig(f'conf_matrix.png')
    plt.close()




