import argparse
import json
import os
import io
import torch
import time
from tqdm import tqdm
import librosa
import os, traceback, sys
from sphfile import SPHFile # pip install sphfile
import numpy as np
import torchmetrics
import csv
from collections import Counter, OrderedDict
import scipy.spatial.distance
import soundfile as sf
import sklearn
import torch.nn.functional as F
from scipy.spatial.distance import cdist
import torchnet as tnt  # pip install git+https://github.com/pytorch/tnt.git@master
from loguru import logger
import matplotlib.pyplot as plt
from sklearn.cluster import SpectralClustering

from modules.dict_to_obj import DictToObj
from modules.file_utils import FileUtils
from modules_core.datasource_v1 import Dataset


def load_model_dir(dir):
    weight_files = [i for i in os.listdir(dir) if i.endswith('.pt')]
    json_files = [i for i in os.listdir(dir) if i.endswith('.json')]
    models_to_use = []
    for json_file in json_files:
        json_filename = json_file.split('.')[0]
        for weight_file in weight_files:
            if json_filename in weight_file:
                models_to_use.append(os.path.join(dir, json_file))
                break

    return models_to_use


def load_file(filename):
    try:
        if filename.endswith(('.json', '.csv', '.eaf', '.xml', '.mmap')):
            data = None
        elif filename.endswith('.sph'):
            sph = SPHFile(filename)
            data = librosa.util.buf_to_float(sph.content, n_bytes=sph.format['sample_n_bytes'])
            if data.shape[0] == 0:
                logger.error(f'no data found inside: {filename}')
                return None
            if sph.format['sample_rate'] != args.sample_rate:
                sr = sph.format['sample_rate']
                logger.info(f'Different sample rate {filename} {sr}')
                data = librosa.core.resample(data, sr, args.sample_rate)
                # for ted talks that are in sph format
                # leave out beginning and end 15sec / 20sec
                data = data[15 * args.sample_rate:(data.shape[0] - args.sample_rate * 20)]
        else:
            data, _ = librosa.load(filename, sr=args.data_samplerate)
        return data
    except Exception as e:
        logger.error(str(e))
        exc_type, exc_value, exc_tb = sys.exc_info()
        logger.error(traceback.format_exception(exc_type, exc_value, exc_tb))


def model_init(path_model_args):
    global args
    with open(path_model_args, 'r') as fp:
        # with io.open(path_model_args, 'r', encoding='utf8') as fp:
        model_args = json.load(fp)
        args = vars(args)
        args.update(model_args)
        args = DictToObj(**args)
        model_args = DictToObj(**model_args)

    path_model = path_model_args.replace('.json', '.pt')

    Model = getattr(__import__('modules_core.' + model_args.model, fromlist=['Model']), 'Model')
    model = Model(model_args)

    if torch.cuda.is_available() and args.is_cuda:
        args.device = 'cuda'
        model = model.cuda()
        model.load_state_dict(torch.load(path_model, map_location=lambda storage, loc: storage.cuda()))
    else:
        args.device = 'cpu'
        model.load_state_dict(torch.load(path_model, map_location='cpu'))

    model = model.eval()
    logger.info('model loaded')
    return model


def model_forward(model, input):
    output_post_np = None
    if type(input) != torch.Tensor:
        input = torch.from_numpy(input)
    if torch.cuda.is_available() and args.is_cuda:
        input = input.cuda()

    with torch.no_grad():
        output = model.forward(input)

    output = output.squeeze(1).to('cpu').detach()
    output_np = output.numpy()

    return output_np


def process_file(model, file):
    input_size = args.input_size
    audio = load_file(file)
    if audio is None:
        return None
    pad_len = input_size - (len(audio) % input_size)
    audio = np.pad(audio, (0, pad_len))
    t_audio = torch.FloatTensor(audio).to(args.device).unsqueeze(dim=0).unsqueeze(dim=0).unsqueeze(dim=0)
    t_audio_strided = F.unfold(t_audio, kernel_size=(1, input_size), stride=(1, input_size)).squeeze().T

    if len(t_audio_strided.shape) == 1:  # e.g. (32000) mdoel will output error. Need to unsqueeze
        t_audio_strided = t_audio_strided.unsqueeze(dim=0)

    loader = torch.utils.data.DataLoader(
        torch.utils.data.TensorDataset(t_audio_strided),
        batch_size=args.batch_size_inf,
        shuffle=False
    )
    embeddings = []
    for batch, in tqdm(loader, desc='voice id'):
        embs = model_forward(model, batch)
        embeddings += embs.tolist()
    return np.array(embeddings)


def main():
    # create output folder
    FileUtils.deleteDir(args.path_output)
    FileUtils.createDir(args.path_output)

    # check if dir to usersid folder or files
    dir_content = os.listdir(args.path_data)
    is_folder_dir = True
    for c in dir_content:
        if os.path.isfile(os.path.join(args.path_data, c)):
            is_folder_dir = False
    # prepare folders to iterate

    if is_folder_dir:
        folders = [os.path.join(args.path_data, i) for i in os.listdir(args.path_data)]
    else:
        folders = [args.path_data]

    # init model
    model = model_init(args.path_model_args)

    # define class names to use
    classes = [
        'happy',
        'sad',
        'angry',
        'other'
    ]
    accuracy_global = torchmetrics.Accuracy()
    results = []
    embeddings_all = []
    labels_all = []
    for folder in tqdm(folders):
        # check data if folder
        classes_in_folder = []
        for file in os.listdir(folder):
            label = None
            for emo in classes:
                if f'emotion_{emo}' in file:
                    label = emo
                    break
            if label is None:
                label = 'other'
            classes_in_folder.append(label)
        if len(set(classes_in_folder)) < args.emotion_count_cutoff:
            continue

        embeddings = []
        labels = []

        for file in os.listdir(folder):
            file_path = os.path.join(folder, file)

            label = None
            for emo in classes:
                if f'emotion_{emo}' in file:
                    label = emo
                    break
            if label is None:
                label = 'other'

            # process voice_id
            embs = process_file(model, file_path)
            if embs is None:
                continue

            for e in embs:
                embeddings.append(e)
                labels.append(classes.index(label))

        labels = np.array(labels)
        labels_all.append(labels)
        embeddings = np.stack(embeddings)
        embeddings_all.append(embeddings)

        # get centers
        centers_by_classes = {}
        for l in range(len(classes)):
            cur_l = labels == l
            if not any(cur_l):
                continue
            cur_emb = embeddings[cur_l]
            # calculate initial center
            centers_by_classes[l] = np.mean(cur_emb, axis=0)

            # remove 90th percentile of samples
            if args.is_90th_percent_centers:
                distance_to_center = cdist(
                    cur_emb,
                    np.expand_dims(centers_by_classes[l], axis=0),
                    metric='euclidean'
                ).squeeze()
                indexes_sorted = np.argsort(distance_to_center)
                embeddings_filtered = cur_emb[indexes_sorted[:int(len(cur_emb)*0.9)]]

                # recalculate centers
                centers_by_classes[l] = np.mean(embeddings_filtered, axis=0)

        # calculate accuracy
        accuracy = torchmetrics.Accuracy()
        confusion_meter = torchmetrics.ConfusionMatrix(num_classes=len(args.classes), normalize="true")
        for idx, emb in enumerate(embeddings):
            dists_to_centers = cdist(
                np.array(list(centers_by_classes.values())),
                np.expand_dims(emb, axis=0),
                metric='cosine'
            ).squeeze()
            idx_closest = np.argmin(dists_to_centers)
            y_prim_idx_each = list(centers_by_classes.keys())[idx_closest]
            # prepare data for metrics
            y_prim_idx_each = F.one_hot(torch.LongTensor((y_prim_idx_each,)), len(args.classes)).type(torch.FloatTensor)
            label = torch.LongTensor((labels[idx],))

            # update local metrics
            accuracy.update(y_prim_idx_each, label)
            confusion_meter.update(y_prim_idx_each, label)
            # update global metrics
            accuracy_global.update(y_prim_idx_each, label)

        logger.info(f'accuracy: {accuracy.compute()}')

        pca = sklearn.decomposition.PCA(n_components=2)
        embeddings_2d = pca.fit_transform(embeddings)
        plt.cla()
        plt.clf()
        scatter = plt.scatter(embeddings_2d[:, -1], embeddings_2d[:, -2], s=1, c=labels, cmap='jet')
        plt.legend(handles=scatter.legend_elements()[0], labels=classes)
        plt.tight_layout(pad=0.5)
        plt.savefig(os.path.join(args.path_output, f'embeddings_{folder.split(os.sep)[-1]}.png'))
        plt.close()
        plt.show()

        class_names = []
        for l in labels:
            class_names.append(classes[l])
        classes_used = ' '.join(set(class_names))
        # write results in dict
        results.append({
            'name': folder.split(os.sep)[-1],
            'classes': classes_used,
            'accuracy': accuracy.compute().item()
        })

    results.append({
        'name': 'All',
        'classes': '-',
        'accuracy': accuracy_global.compute().item()
    })

    embeddings_all = np.vstack(embeddings_all)
    labels_all = np.hstack(labels_all)

    # get centers
    centers_by_classes = {}
    for l in range(len(classes)):
        cur_l = labels_all == l
        if not any(cur_l):
            continue
        cur_emb = embeddings_all[cur_l]
        # calculate initial center
        centers_by_classes[l] = np.mean(cur_emb, axis=0)

        # remove 90th percentile of samples
        if args.is_90th_percent_centers:
            distance_to_center = cdist(
                cur_emb,
                np.expand_dims(centers_by_classes[l], axis=0),
                metric='euclidean'
            ).squeeze()
            indexes_sorted = np.argsort(distance_to_center)
            embeddings_filtered = cur_emb[indexes_sorted[:int(len(cur_emb)*0.9)]]

            # recalculate centers
            centers_by_classes[l] = np.mean(embeddings_filtered, axis=0)

    # calculate accuracy
    accuracy = torchmetrics.Accuracy()
    confusion_meter = torchmetrics.ConfusionMatrix(num_classes=len(args.classes), normalize="true")
    for idx, emb in enumerate(embeddings_all):
        dists_to_centers = cdist(
            np.array(list(centers_by_classes.values())),
            np.expand_dims(emb, axis=0),
            metric='cosine'
        ).squeeze()
        idx_closest = np.argmin(dists_to_centers)
        y_prim_idx_each = list(centers_by_classes.keys())[idx_closest]
        # prepare data for metrics
        y_prim_idx_each = F.one_hot(torch.LongTensor((y_prim_idx_each,)), len(args.classes)).type(torch.FloatTensor)
        label = torch.LongTensor((labels_all[idx],))

        # update local metrics
        accuracy.update(y_prim_idx_each, label)
        confusion_meter.update(y_prim_idx_each, label)

    logger.info(f'Total accuracy: {accuracy.compute()}')

    pca = sklearn.decomposition.PCA(n_components=2)
    embeddings_2d = pca.fit_transform(embeddings_all)
    plt.cla()
    plt.clf()
    scatter = plt.scatter(embeddings_2d[:, -1], embeddings_2d[:, -2], s=1, c=labels_all, cmap='jet')
    plt.legend(handles=scatter.legend_elements()[0], labels=classes)
    plt.tight_layout(pad=0.5)
    plt.savefig(os.path.join(args.path_output, f'embeddings_all.png'))
    plt.close()
    plt.show()

    csv_path = os.path.join(args.path_output, 'results.csv')
    with open(csv_path, 'w', newline='') as output_file:
        dict_writer = csv.DictWriter(output_file, results[0].keys())
        dict_writer.writeheader()
        dict_writer.writerows(results)


if __name__ == '__main__':
    arg_parser = argparse.ArgumentParser()
    arg_parser.add_argument('-path_data', default="./data/voiceid_files", type=str)
    arg_parser.add_argument('-path_output', default="./inference_results", type=str)
    arg_parser.add_argument('-batch_size_inf', default=32, type=int)
    arg_parser.add_argument('-is_cuda', default=True, type=bool)  # type=lambda x: (str(x).lower() == 'true'))
    arg_parser.add_argument('-is_90th_percent_centers', default=True, type=bool)
    arg_parser.add_argument('-emotion_count_cutoff', default=3, type=int)
    arg_parser.add_argument(
        '-path_model_args',
        default='./results/embedding_models/4_class_augmentations_resnet_triplet_continuing-26/4_class_augmentations_resnet_triplet_continuing-26.json',
        type=str
    )

    args, args_other = arg_parser.parse_known_args()

    FileUtils.createDir('./logs')
    logger.add('./logs/inference-{time:YYYY-MM-DD}.log', rotation='00:00')

    logger.info('Inference script starting')
    logger.info(json.dumps(args.__dict__, indent=4))

    main()

    logger.info('Inference script ending')
