import argparse
import json
import os
import io
import torch
import time
from tqdm import tqdm
import librosa
import os, traceback, sys
from matplotlib.backends.backend_agg import FigureCanvasAgg as FigureCanvas
from sphfile import SPHFile # pip install sphfile
import numpy as np
import torchmetrics
import re
import sklearn
import torch.nn.functional as F
from scipy.spatial.distance import cdist
from loguru import logger
import matplotlib.pyplot as plt

from modules.dict_to_obj import DictToObj
from modules.file_utils import FileUtils


def save_confusion_matrix(dataXY, ticks, is_percent_added=True):
    fig = plt.figure()
    plt.clf()
    ax = fig.add_subplot(111)
    ax.set_aspect(1)
    res = ax.imshow(dataXY, cmap=plt.cm.Blues if is_percent_added else plt.cm.binary,
                    interpolation='nearest')

    width, height = dataXY.shape

    if is_percent_added:
        cmap_min, cmap_max = res.cmap(0), res.cmap(256)
        thresh = (dataXY.max() + dataXY.min()) / 2.0
        for x in range(width):
            for y in range(height):
                color = cmap_max if dataXY[x][y] < thresh else cmap_min
                ax.annotate(str(round(float(dataXY[x][y]), 3)), xy=(y, x),  # percent without decimals
                            horizontalalignment='center',
                            verticalalignment='center',
                            color=color)

    cb = fig.colorbar(res)
    if ticks is not None:
        plt.xticks(range(width), ticks, rotation=90)
        plt.yticks(range(height), ticks)

    plt.xlabel('Predicted class')
    plt.ylabel('Actual class')

    fig.set_tight_layout(True)
    canvas = FigureCanvas(fig)
    canvas.draw()
    plt.savefig(os.path.join(args.path_output, f'confusion_matrix.png'))
    plt.close(fig)


def load_model_dir(dir):
    weight_files = [i for i in os.listdir(dir) if i.endswith('.pt')]
    json_files = [i for i in os.listdir(dir) if i.endswith('.json')]
    models_to_use = []
    for json_file in json_files:
        json_filename = json_file.split('.')[0]
        for weight_file in weight_files:
            if json_filename in weight_file:
                models_to_use.append(os.path.join(dir, json_file))
                break

    return models_to_use


def load_file(filename):
    try:
        if filename.endswith(('.json', '.csv', '.eaf', '.xml', '.mmap')):
            data = None
        elif filename.endswith('.sph'):
            sph = SPHFile(filename)
            data = librosa.util.buf_to_float(sph.content, n_bytes=sph.format['sample_n_bytes'])
            if data.shape[0] == 0:
                logger.error(f'no data found inside: {filename}')
                return None
            if sph.format['sample_rate'] != args.sample_rate:
                sr = sph.format['sample_rate']
                logger.info(f'Different sample rate {filename} {sr}')
                data = librosa.core.resample(data, sr, args.sample_rate)
                # for ted talks that are in sph format
                # leave out beginning and end 15sec / 20sec
                data = data[15 * args.sample_rate:(data.shape[0] - args.sample_rate * 20)]
        else:
            data, _ = librosa.load(filename, sr=args.data_samplerate)
        return data
    except Exception as e:
        logger.error(str(e))
        exc_type, exc_value, exc_tb = sys.exc_info()
        logger.error(traceback.format_exception(exc_type, exc_value, exc_tb))


def model_init(path_model_args):
    global args
    with open(path_model_args, 'r') as fp:
        # with io.open(path_model_args, 'r', encoding='utf8') as fp:
        model_args = json.load(fp)
        args = vars(args)
        args.update(model_args)
        args = DictToObj(**args)
        model_args = DictToObj(**model_args)

    path_model = path_model_args.replace('.json', '.pt')
    # add missing args (TODO fix later)
    model_args.device = 'cpu'
    # model_args.input_size =

    Model = getattr(__import__('modules_core.' + model_args.model, fromlist=['Model']), 'Model')
    model = Model(model_args)

    if torch.cuda.is_available() and args.is_cuda:
        args.device = 'cuda'
        model = model.cuda()
        model.load_state_dict(torch.load(path_model, map_location=lambda storage, loc: storage.cuda()))
    else:
        args.device = 'cpu'
        model.load_state_dict(torch.load(path_model, map_location='cpu'))

    model = model.eval()
    logger.info('model loaded')
    return model


def model_forward(model, input):
    output_post_np = None
    if type(input) != torch.Tensor:
        input = torch.from_numpy(input)
    if torch.cuda.is_available() and args.is_cuda:
        input = input.cuda()

    with torch.no_grad():
        embedding_loss, x_dec, x_in, perplexity, z_q = model.forward(input)

    output = z_q.unsqueeze(dim=0).to('cpu').detach()
    output_np = output.numpy()

    return output_np


def process_file(model, file):
    audio = load_file(file)
    if audio is None:
        return None
    input_size = args.input_size
    step_size = int(args.input_size / 2)
    pad_reference = step_size
    if len(audio) < input_size:
        pad_reference = input_size
    overflow = len(audio) % pad_reference
    if overflow > 0:
        pad_len = pad_reference - overflow
        audio = np.pad(audio, (0, pad_len))
    t_audio = torch.FloatTensor(audio).to(args.device).unsqueeze(dim=0).unsqueeze(dim=0).unsqueeze(dim=0)
    t_audio_strided = F.unfold(t_audio, kernel_size=(1, input_size), stride=(1, input_size)).squeeze().T

    if len(t_audio_strided.shape) == 1:  # e.g. (32000) mdoel will output error. Need to unsqueeze
        t_audio_strided = t_audio_strided.unsqueeze(dim=0)

    loader = torch.utils.data.DataLoader(
        torch.utils.data.TensorDataset(t_audio_strided),
        batch_size=args.batch_size_inf,
        shuffle=False
    )
    embeddings = []
    for batch, in tqdm(loader, desc='voice id'):
        embs = model_forward(model, batch)
        embeddings += embs.tolist()
    return np.array(embeddings)


def main():
    # create output folder
    FileUtils.deleteDir(args.path_output)
    FileUtils.createDir(args.path_output)

    model = model_init(args.path_model_args)

    classes = model.args.classes
    embeddings = []
    labels = []
    for file in tqdm(os.listdir(args.path_data)):
        file_path = os.path.join(args.path_data, file)

        if '.wav' not in file:
            continue

        label = re.search("happiness|sadness|anger|neutral|other", file)[0]
        if label not in args.classes:
            continue

        # if args.gender_inf:
        #     gender = re.search("male|female", file)[0]
        #     if gender != args.gender_inf:
        #         continue

        audio = load_file(file_path)
        audio_t = torch.FloatTensor(audio).unsqueeze(0)
        embs = model_forward(model, audio_t)
        if embs is None:
            continue

        for e in embs:
            for emo in classes:
                if emo in file:
                    embeddings.append(e)
                    labels.append(classes.index(emo))
                    break

    labels = np.array(labels)
    embeddings = np.stack(embeddings)

    # get centers
    centers_by_classes = {}
    for l in range(len(classes)):
        cur_l = labels == l
        cur_emb = embeddings[cur_l]
        # calculate initial center
        centers_by_classes[l] = np.mean(cur_emb, axis=0)

        # remove 90th percentile of samples
        if args.is_90th_percent_centers:
            distance_to_center = cdist(
                cur_emb,
                np.expand_dims(centers_by_classes[l], axis=0),
                metric='euclidean'
            ).squeeze()
            indexes_sorted = np.argsort(distance_to_center)
            embeddings_filtered = cur_emb[indexes_sorted[:int(len(cur_emb)*0.9)]]

            # recalculate centers
            centers_by_classes[l] = np.mean(embeddings_filtered, axis=0)

    # calculate accuracy
    accuracy = torchmetrics.Accuracy()
    confusion_meter = torchmetrics.ConfusionMatrix(num_classes=len(args.classes), normalize="true")
    for idx, emb in enumerate(embeddings):
        dists_to_centers = cdist(
            np.array(list(centers_by_classes.values())),
            np.expand_dims(emb, axis=0),
            metric='euclidean'
        ).squeeze()
        idx_closest = np.argmin(dists_to_centers)
        y_prim_idx_each = list(centers_by_classes.keys())[idx_closest]
        # prepare data for metrics
        y_prim_idx_each = F.one_hot(torch.LongTensor((y_prim_idx_each,)), len(args.classes)).type(torch.FloatTensor)
        label = torch.LongTensor((labels[idx],))

        accuracy.update(y_prim_idx_each, label)
        confusion_meter.update(y_prim_idx_each, label)

    logger.info(f'accuracy: {accuracy.compute()}')

    save_confusion_matrix(confusion_meter.compute(), classes)

    pca = sklearn.decomposition.PCA(n_components=2)
    embeddings_2d = pca.fit_transform(embeddings)
    plt.cla()
    plt.clf()
    scatter = plt.scatter(embeddings_2d[:, -1], embeddings_2d[:, -2], s=1, c=labels, cmap='jet')
    plt.legend(handles=scatter.legend_elements()[0], labels=classes)
    plt.tight_layout(pad=0.5)
    plt.savefig(os.path.join(args.path_output, f'embeddings.png'))
    plt.close()


if __name__ == '__main__':
    arg_parser = argparse.ArgumentParser()
    arg_parser.add_argument('-path_data', default="./inference_and_test_data/emo_relabeled_data_raw_improv_team_2021_4sec_v1", type=str)
    arg_parser.add_argument('-path_output', default="./inference_results", type=str)
    arg_parser.add_argument('-gender_inf', default="male", type=str)
    arg_parser.add_argument('-batch_size_inf', default=32, type=int)
    arg_parser.add_argument('-is_cuda', default=False, type=bool)  # type=lambda x: (str(x).lower() == 'true'))
    arg_parser.add_argument('-is_90th_percent_centers', default=True, type=bool)
    arg_parser.add_argument(
        '-path_model_args',
        default='./results/best_accuracy.json',
        type=str
    )

    args, args_other = arg_parser.parse_known_args()

    FileUtils.createDir('./logs')
    logger.add('./logs/inference-{time:YYYY-MM-DD}.log', rotation='00:00')

    logger.info('Inference script starting')
    logger.info(json.dumps(args.__dict__, indent=4))

    main()

    logger.info('Inference script ending')
