import argparse
import copy
import math
import datetime
import os
import time
import torchmetrics
import librosa
import re

from loguru import logger
from tqdm import tqdm

from modules.audio_augmentations import *
from modules.args_utils import ArgsUtils
from modules.csv_utils_2 import CsvUtils2
from modules.dict_to_obj import DictToObj
from modules.file_utils import FileUtils
from modules.metric_utils import LossMeter
from modules.tensorboard_utils import TensorBoardUtils, CustomSummaryWriter
from modules.main_helpers import get_param_count
from modules.demucs import Demucs, DemucsStreamer

if __name__ == '__main__':

    parser = argparse.ArgumentParser(add_help=False)

    # Dir creation params
    parser.add_argument('-id', default=0, type=int)
    parser.add_argument('-sequence_name', default='sequence', type=str)
    parser.add_argument('-run_name', default='run', type=str)

    # Script imports
    parser.add_argument('-model', default='model_1D_classification_demucs_v1', type=str)
    parser.add_argument('-datasource', default='datasource_v1', type=str)
    parser.add_argument('-is_single_cuda_device', default=True, type=lambda x: (str(x).lower() == 'true'))

    # Datasource params
    parser.add_argument('-path_train_emo', default='./inference_and_test_data/emo_audio_relabelled_threshold_3_test_train_split_v3/test_dataset.json', type=str)
    parser.add_argument('-path_test_emo', default='./inference_and_test_data/emo_audio_relabelled_threshold_3_test_train_split_v3/test_dataset.json', type=str)
    parser.add_argument('-path_validation_improv',
                        default='./inference_and_test_data/emo_relabeled_data_raw_improv_team_2021_4sec_v1', type=str)
    parser.add_argument('-path_validation_PP',
                        default='./inference_and_test_data/emo_relabeled_data_raw_improv_team_2021_4sec_v1', type=str)
    parser.add_argument('-data_workers', default=0, type=int)
    parser.add_argument('-data_samplerate', default=16000, type=int)  # Samples are in 16k freq
    parser.add_argument('-data_scaling_type', default='standardization', type=str)  # [standardization; normalization; none]
    parser.add_argument('-audio_cutoff', default=100.0, type=float)

    # Dataloader params
    parser.add_argument('-downsample_input', default=True, type=lambda x: (str(x).lower() == 'true'))
    # downsamples input to 8k and upsamples it back to 16k
    parser.add_argument('-method', default='division', type=str)
    parser.add_argument('-alpha', default=1.0, type=float)

    # Audio augmentation params
    parser.add_argument('-each_augment_proba', default=0.0, type=float)
    parser.add_argument('-is_shift_augmentation', default=False, type=lambda x: (str(x).lower() == 'true'))
    parser.add_argument('-shift_size', default=512, type=int)

    parser.add_argument('-is_bandmask_augmentation', default=False, type=lambda x: (str(x).lower() == 'true'))
    parser.add_argument('-bandmask_maxwidth', default=0.2, type=float)
    parser.add_argument('-bandmask_bands', default=120, type=int)

    parser.add_argument('-is_revecho_augmentation', default=False, type=lambda x: (str(x).lower() == 'true'))

    parser.add_argument('-use_denoising', default=False, type=lambda x: (str(x).lower() == 'true'))
    parser.add_argument('-denoising_probability', default=0.5, type=float)
    parser.add_argument('-path_denoising_model_args', default=r'/Users/aigaandrijanova/Downloads/facebook_pretrained_dns64.th', type=str)
    parser.add_argument('-mix_dry_part', default=0.05, type=float)

    # Model params
    parser.add_argument('-apply_normalization', default=True, type=lambda x: (str(x).lower() == 'true'))
    parser.add_argument('-normalization_scope', default='sample', type=str)  # [sample; batch; dataset]
    parser.add_argument('-encoder_ch_out', default=16, type=int)
    parser.add_argument('-resample', default=1, type=int)
    parser.add_argument('-model_mean', default=0, type=float)
    parser.add_argument('-model_std', default=0, type=float)
    parser.add_argument('-normalization_module', default='instance', type=str)  # [batch; instance; none]
    parser.add_argument('-input_size', default=64_000, type=int)  # used in inference

    # Demucs encoder params
    parser.add_argument('-encoder_depth', default=2, type=int)  # 5
    parser.add_argument('-encoder_hidden_size', default=16, type=int)  # 48
    parser.add_argument('-encoder_kernel_size', default=8, type=int)
    parser.add_argument('-encoder_stride', default=4, type=int)
    parser.add_argument('-encoder_growth', default=2, type=int)

    # Spectrogram params
    parser.add_argument('-spectrogram_type', default='melspectrogram', type=str)  # [spectrogram; melspectrogram; mfcc]
    parser.add_argument('-n_fft', default=1024, type=int)
    parser.add_argument('-win_length', default=512, type=int)
    parser.add_argument('-hop_length', default=128, type=int)
    parser.add_argument('-n_mels', default=128, type=int)
    parser.add_argument('-n_mfcc', default=64, type=int)

    # Main params
    parser.add_argument('-epochs', default=5, type=int)
    parser.add_argument('-batch_size', default=16, type=int)
    parser.add_argument('-optimizer', default='adam', type=str)
    parser.add_argument('-learning_rate', default=3e-4, type=float)

    # Metric plot and model saving params
    parser.add_argument('-save_models', default=True, type=lambda x: (str(x).lower() == 'true'))
    parser.add_argument('-inference_gender', default='any', type=str)  # [any; female; male]

    # Early stopping params
    parser.add_argument('-early_stopping_patience', default=2, type=int)
    parser.add_argument('-early_stopping_param', default='test_loss', type=str)
    parser.add_argument('-early_stopping_delta_percent', default=3e-3, type=float)
    parser.add_argument('-early_stopping_param_coef', default=-1.0, type=float)

    parser.add_argument('-classes', default=[
        'happiness',
        'anger',
        'sadness',
        'other'
    ], nargs='*')

    def deserialize_model(package, strict=False):
        """deserialize_model.

        """
        if 'class' in package:
            klass = package['class']
            if strict:
                model_d = klass(*package['args'], **package['kwargs'])
            else:
                sig = inspect.signature(klass)
                kw = package['kwargs']
                for key in list(kw):
                    if key not in sig.parameters:
                        del kw[key]
                model_d = klass(*package['args'], **kw)
            model_d.load_state_dict(package['state'])
        else:
            model_d = Demucs(hidden=64)
            model_d.load_state_dict(package)

        return model_d

    def model_init(path_model_args):
        pkg = torch.load(path_model_args, 'cpu')
        model = deserialize_model(pkg)

        if torch.cuda.is_available():
            args.device = 'cuda'
            model = model.cuda()
        else:
            args.device = 'cpu'

        model = model.eval()

        return model

    args, args_other = parser.parse_known_args()
    args = ArgsUtils.add_other_args_v2(args, args_other)

    args_dict_orig = copy.deepcopy(args.__dict__)
    args_dict = args.__dict__
    args = DictToObj(**args_dict)

    path_sequence = f'./results/{args.sequence_name}'
    args.run_name += ('-' + datetime.datetime.utcnow().strftime(f'%y-%m-%d--%H-%M-%S'))
    path_run = f'./results/{args.sequence_name}/{args.run_name}'
    path_artificats = f'./artifacts/{args.sequence_name}/{args.run_name}'
    FileUtils.createDir(path_run)
    FileUtils.createDir(path_artificats)
    FileUtils.writeJSON(f'{path_run}/args.json', vars(args))

    CsvUtils2.create_global(path_sequence)
    CsvUtils2.create_local(path_sequence, args.run_name)

    base_name = os.path.basename(path_sequence)
    logger.add(f'{path_run}/log-{base_name}.txt')

    tensorboard_writer = CustomSummaryWriter(logdir=path_run)
    tensorboard_utils = TensorBoardUtils(tensorboard_writer)

    args.device = 'cuda'
    if not torch.cuda.is_available():
        args.device = 'cpu'
        logger.info('CUDA NOT AVAILABLE')
    else:
        logger.info('cuda devices: {}'.format(torch.cuda.device_count()))

    loss_func = torch.nn.CrossEntropyLoss(reduction="mean")

    Model = getattr(__import__('modules_core.' + args.model, fromlist=['Model']), 'Model')

    logger.info(f'Loading model')
    if not args.is_single_cuda_device:
        if args.device == 'cuda' and torch.cuda.device_count() > 1:
            model = torch.nn.DataParallel(
                Model(args),
                dim=0,
                device_ids=list(range(torch.cuda.device_count()))).to(args.device)
            logger.info(f'PARALLEL MODEL {torch.cuda.device_count()}')
    else:
        model = Model(args).to(args.device)

    logger.info(f'Model param count: {get_param_count(model)}')

    get_data_loaders = getattr(__import__('modules_core.' + args.datasource, fromlist=['get_data_loaders']),
                               'get_data_loaders')
    data_loader_train, data_loader_test = get_data_loaders(args)

    optimizer_func = None
    if args.optimizer == 'adam':
        optimizer_func = torch.optim.Adam(
            model.parameters(),
            lr=args.learning_rate
        )
    elif args.optimizer == 'radam':
        optimizer_func = torch.optim.RAdam(
            model.parameters(),
            lr=args.learning_rate
        )

    augmentations = []
    if args.is_shift_augmentation:
        augmentations.append(Shift(proba=args.each_augment_proba, shift=args.shift_size))
    if args.is_bandmask_augmentation:
        augmentations.append(BandMask(proba=args.each_augment_proba,
                                      maxwidth=args.bandmask_maxwidth,
                                      bands=args.bandmask_bands,
                                      sample_rate=args.data_samplerate))
    if args.is_revecho_augmentation:
        augmentations.append(RevEcho(proba=args.each_augment_proba, sample_rate=args.data_samplerate))
    augmentations = torch.nn.Sequential(*augmentations)

    if args.use_denoising:
        denoising_model_name = args.path_denoising_model_args.split('.')[0].split(os.sep)[-1]
        denoising_model = model_init(args.path_denoising_model_args)

    def torch_normalization(raw_x):
        min_x = torch.min(raw_x, dim=-1, keepdim=True)
        max_x = torch.max(raw_x, dim=-1, keepdim=True)
        if any(torch.eq(min_x.values,max_x.values)): # check if none of min and max values are equal
            return torch.zeros_like(raw_x)
        raw_x = (raw_x - (max_x.values + min_x.values) * 0.5) / ((max_x.values - min_x.values) * 0.5) #range between -1 and 1
        return raw_x

    def forward_batch(batch, mode):
        x = batch[0].to(args.device)
        y = batch[1].to(args.device)

        if len(augmentations) and mode == 'train':
            for idx in range(len(x)):
                x[idx] = augmentations.forward(x[idx].unsqueeze(dim=0)).squeeze(dim=0)

        if args.use_denoising and mode == 'train':
            if np.random.random() <= args.denoising_probability:
                with torch.no_grad():
                    denoised_samples = denoising_model.forward(x).detach().squeeze(dim=1)
                    denoised_samples = x * args.mix_dry_part + (1 - args.mix_dry_part) * denoised_samples
                    x = torch_normalization(denoised_samples)
                    del denoised_samples

        output = model.forward(x)

        return output, y

    def evaluate_epoch(data_loader, mode):
        with torch.no_grad():
            with tqdm(data_loader, desc=f'{args.run_name} Evaluate {mode} epoch: {epoch}') as t:
                for batch in t:
                    output, y = forward_batch(batch, mode)

                    # prepare bce output for meters
                    if output.size(-1) == 1:
                        output = torch.hstack((output, 1 - output))
                        y = 1 - y

                    output = output.cpu()
                    y = y.cpu()
                    metrics[f'{mode}_accuracy'](output, y.squeeze())
                    metrics[f'{mode}_f1'](output, y.squeeze())
                    metrics[f'{mode}_confusion_meter'](output, y.squeeze())

            state[f'{mode}_accuracy'] = metrics[f'{mode}_accuracy'].compute().item()
            state[f'{mode}_f1'] = metrics[f'{mode}_f1'].compute().item()
            tensorboard_utils.addPlotConfusionMatrix(
                dataXY=metrics[f'{mode}_confusion_meter'].compute().numpy(),
                ticks=args.classes,
                tag=f'{mode}_confusion_matrix',
                global_step=epoch,
                is_percent_added=True
            )


    def forward_epoch(data_loader, mode):
        with tqdm(data_loader, desc=f'{args.run_name} Forward {mode} epoch: {epoch}') as t:
            for batch in t:
                t.set_postfix(loss=metrics[f'{mode}_loss'].compute().item())
                if mode == 'train':
                    optimizer_func.zero_grad()
                    model.zero_grad()

                output, y = forward_batch(batch, mode)

                # calculate loss
                loss = loss_func.forward(output.squeeze(), y.squeeze())

                if mode == 'train':
                    loss.backward()
                    optimizer_func.step()

                loss = loss.to('cpu')
                metrics[f'{mode}_loss'](loss)

            state[f'{mode}_loss'] = metrics[f'{mode}_loss'].compute().item()

    def update_early_stopping():
        # early stopping
        percent_improvement = 0
        early_stopping = False
        if epoch > 1:
            if state_before[args.early_stopping_param] != 0:
                percent_improvement = -(state[args.early_stopping_param] - state_before[args.early_stopping_param]) / \
                                      state_before[args.early_stopping_param]
            if state[args.early_stopping_param] >= 0:
                if args.early_stopping_delta_percent > percent_improvement:
                    state['early_stopping_patience'] += 1
                else:
                    state['early_stopping_patience'] = 0
            state['early_percent_improvement'] = percent_improvement
        if state['early_stopping_patience'] >= args.early_stopping_patience or \
                math.isnan(percent_improvement):
            logger.info('early stopping')
            early_stopping = True

        return percent_improvement, early_stopping

    state = {
        'epoch': 0,
        'best_loss': -1,
        'best_accuracy': -1,
        'best_f1': -1,
        'avg_epoch_time': -1,
        'epoch_time': -1,
        'early_stopping_patience': 0,
        'early_percent_improvement': 0,
        'train_loss': -1,
        'test_loss': -1,
        'train_accuracy': -1,
        'test_accuracy': -1,
        'train_f1': -1,
        'test_f1': -1,
        'eval_accuracy_improv': 0,
        'eval_f1_improv': 0,
        'eval_accuracy_pp': 0,
        'eval_f1_pp': 0
    }

    avg_time_epochs = []
    time_epoch = time.time()

    metrics = dict(
        train_loss=LossMeter(),
        test_loss=LossMeter(),
        train_accuracy=torchmetrics.Accuracy(),
        test_accuracy=torchmetrics.Accuracy(),
        train_f1=torchmetrics.F1Score(),
        test_f1=torchmetrics.F1Score(),
        train_confusion_meter=torchmetrics.ConfusionMatrix(num_classes=len(args.classes), normalize="true"),
        test_confusion_meter=torchmetrics.ConfusionMatrix(num_classes=len(args.classes), normalize="true")
    )

    best_checkpoint = {}
    for epoch in range(0, args.epochs + 1):
        state_before = copy.deepcopy(state)
        logger.info('epoch: {} / {}'.format(epoch, args.epochs))

        for key in metrics.keys():
            metrics[key].reset()

        # training phase
        # train model
        meter_prefix = 'train'
        model = model.train()
        forward_epoch(data_loader_train, meter_prefix)
        
        # evaluation phase
        with torch.no_grad():
            model = model.eval()
            # test set forward
            meter_prefix = 'test'
            forward_epoch(data_loader_test, meter_prefix)
            evaluate_epoch(data_loader_test, meter_prefix)
            percent_improvement, early_stopping = update_early_stopping()

        model_module = model
        if not args.is_single_cuda_device:
            model_module = model.module

        # save best metrics
        if epoch == 0:
            state['best_loss'] = state['test_loss']
        elif state['best_loss'] > state['test_loss']:
            state['best_loss'] = state['test_loss']
            if args.save_models:
                torch.save(model_module.state_dict(), os.path.join(path_run, 'best_loss.pt'))
                # save checkpoint
                best_checkpoint['model'] = model_module.state_dict()
                best_checkpoint['optimizer'] = optimizer_func.state_dict()
                best_checkpoint['state'] = state
                best_checkpoint['args'] = args
                torch.save(best_checkpoint, os.path.join(path_run, 'best_checkpoint.pt'))

        if epoch == 0:
            state['best_accuracy'] = state['test_accuracy']
        elif state['best_accuracy'] < state['test_accuracy']:
            state['best_accuracy'] = state['test_accuracy']
            if args.save_models:
                torch.save(model_module.state_dict(), os.path.join(path_run, 'best_accuracy.pt'))

        if epoch == 0:
            state['best_f1'] = state['test_f1']
        elif state['best_f1'] < state['test_f1']:
            state['best_f1'] = state['test_f1']
            if args.save_models:
                torch.save(model_module.state_dict(), os.path.join(path_run, 'best_f1.pt'))

        epoch_time = (time.time() - time_epoch) / 60.0
        percent = epoch / args.epochs
        state['epoch_time'] = epoch_time

        avg_time_epochs.append(epoch_time)
        state['avg_epoch_time'] = np.average(avg_time_epochs)
        eta = ((args.epochs - epoch) * state['avg_epoch_time'])
        time_epoch = time.time()
        state['epoch'] = epoch

        # save last model
        torch.save(model_module.state_dict(), os.path.join(path_run, f'{args.run_name}.pt'))

        # log state each epoch
        tensorboard_writer.add_hparams(vars(args), state, name=args.run_name, global_step=epoch)

        CsvUtils2.add_hparams(
            path_sequence=path_sequence,
            run_name=args.run_name,
            args_dict=args.__dict__,
            metrics_dict=state,
            global_step=epoch
        )

        logger.info(
            f'{round(percent * 100, 2)}% each: {round(state["avg_epoch_time"], 2)}min, eta: {round(eta, 2)} min, train loss: {round(state["train_loss"], 5)}, '
            f'test loss: {round(state["test_loss"], 5)}, improve: {round(percent_improvement, 3)}, train acc: {round(state["train_accuracy"], 2)}, '
            f'test acc: {round(state["test_accuracy"], 2)}, train f1: {round(state["train_f1"], 2)}, test f1: {round(state["test_f1"], 2)}')

        if early_stopping:
            break

    # run eval
    model.load_state_dict(torch.load(os.path.join(path_run, f'best_accuracy.pt'),
                                     map_location=lambda storage, loc: storage.cuda()))
    model = model.eval()

    for validation_set in ['improv', 'pp']:
        eval_accuracy = torchmetrics.Accuracy(average='macro')
        eval_f1 = torchmetrics.F1Score(average='macro')
        eval_confusion_meter = torchmetrics.ConfusionMatrix(num_classes=len(args.classes), normalize="true")

        if validation_set == 'improv':
            path = args.path_validation_improv
        elif validation_set == 'pp':
            path = args.path_validation_pp
        samples = [it for it in os.listdir(args.path)]

        pred_labels = []
        true_labels = []
        with torch.no_grad():
            for idx, name in enumerate(tqdm(samples, desc='Evaluate model')):
                if '.wav' not in name:
                    continue
                audio, _ = librosa.load(os.path.join(args.path_validation, name), sr=args.data_samplerate)
                if audio is None:
                    break
                else:
                    label = re.search("happiness|sadness|anger|neutral|other", name)[0]
                    if label == 'neutral':
                        label = 'other'
                    if label not in args.classes:
                        continue

                    gender = re.search("female|male", name)[0]
                    if not(args.inference_gender == 'any' or args.inference_gender == gender):
                        continue

                input_size = args.input_size
                step_size = int(args.input_size / 2)
                pad_reference = step_size
                if len(audio) < input_size:
                    pad_reference = input_size
                overflow = len(audio) % pad_reference
                if overflow > 0:
                    pad_len = pad_reference - overflow
                    audio = np.pad(audio, (0, pad_len))
                t_audio = torch.FloatTensor(audio).to(args.device).unsqueeze(dim=0).unsqueeze(dim=0).unsqueeze(dim=0)
                t_audio_strided = F.unfold(t_audio, kernel_size=(1, input_size), stride=(1, step_size)).squeeze().T

                if len(t_audio_strided.shape) == 1:
                    t_audio_strided = t_audio_strided.unsqueeze(dim=0)

                loader = torch.utils.data.DataLoader(
                    torch.utils.data.TensorDataset(t_audio_strided),
                    batch_size=args.batch_size,
                    shuffle=False
                )
                for batch, in loader:
                    result = model.forward(batch).cpu()
                    label = torch.as_tensor([args.classes.index(label)])
                    eval_accuracy.update(result, label)
                    eval_f1.update(result, label)
                    eval_confusion_meter.update(result, label)

                if validation_set == 'improv':
                    state['eval_accuracy_improv'] = eval_accuracy.compute().item()
                    state['eval_f1_improv'] = eval_f1.compute().item()
                elif validation_set == 'pp':
                    state['eval_accuracy_pp'] = eval_accuracy.compute().item()
                    state['eval_f1_pp'] = eval_f1.compute().item()

    tensorboard_writer.add_hparams(vars(args), state, name=args.run_name, global_step=epoch)

    CsvUtils2.add_hparams(
        path_sequence=path_sequence,
        run_name=args.run_name,
        args_dict=args.__dict__,
        metrics_dict=state,
        global_step=epoch
    )

    tensorboard_utils.addPlotConfusionMatrix(
        dataXY=eval_confusion_meter.compute().numpy(),
        ticks=args.classes,
        tag=f'eval_confusion_matrix',
        global_step=epoch,
        is_percent_added=True
    )
    logger.info(f'Evaluate results; eval_accuracy: {eval_accuracy.compute()}, eval_f1: {eval_f1.compute()}')


    tensorboard_writer.close()
