import torch
import torch.nn
import torch.utils.data
import torch.utils.data.sampler
import numpy as np

from loguru import logger
from modules.file_utils import FileUtils


class Dataset(torch.utils.data.dataset.Dataset):
    def __init__(self, args, datasource_json):
        super().__init__()
        self.args = args

        self.json_desc = FileUtils.readJSON(datasource_json)

        self.mmap_x = np.memmap(
            datasource_json.replace('.json', '.mmap'),
            dtype=np.float16,
            shape=tuple(self.json_desc['shape'])
        )

        self.lengths = self.json_desc['lengths']

        self.genders = self.json_desc['genders']
        self.sample_ids = self.json_desc['sample_ids']
        self.speaker_ids = self.json_desc['speaker_ids']
        self.data_samplerate = self.args.data_samplerate

        labels = self.json_desc['labels']
        self.y_str = [labels[i] for i in self.json_desc['class_idxes']]
        self.classes = self.json_desc['classes']

        # # Filter emotions
        # if len(self.args.exclude_classes) > 0:
        #     self.classes = [e for e in self.classes if e not in self.args.exclude_classes]
        #
        #     lengths = []
        #     genders = []
        #     sample_ids = []
        #     speaker_ids = []
        #     y_str = []
        #     for idx, l in enumerate(self.y_str):
        #         if l in self.classes:
        #             lengths.append(self.lengths[idx])
        #             genders.append(self.genders[idx])
        #             sample_ids.append(self.sample_ids[idx])
        #             speaker_ids.append(self.speaker_ids[idx])
        #             y_str.append(self.y_str[idx])
        #     self.lengths = lengths
        #     self.genders = genders
        #     self.sample_ids = sample_ids
        #     self.speaker_ids = speaker_ids
        #     self.y_str = y_str
        #
        # Filter datasets
        lengths = []
        genders = []
        sample_ids = []
        speaker_ids = []
        y_str = []
        for idx, s_id in enumerate(self.sample_ids):
            dataset_id = int(s_id.split("_")[1])
            if self.json_desc["dataset_names"][dataset_id] in self.args.include_datasets:
                lengths.append(self.lengths[idx])
                genders.append(self.genders[idx])
                sample_ids.append(self.sample_ids[idx])
                speaker_ids.append(self.speaker_ids[idx])
                y_str.append(self.y_str[idx])
        self.lengths = lengths
        self.genders = genders
        self.sample_ids = sample_ids
        self.speaker_ids = speaker_ids
        self.y_str = y_str

    def __len__(self):
        return len(self.lengths)

    def __getitem__(self, idx):
        y_str = self.y_str[idx]
        y = self.args.classes.index(y_str)
        x = self.mmap_x[idx]  # np.float16 (min eps = 0.000977, not enough for normalization)
        x = x.astype(np.float32)  # np.float32 (min eps = 1.1920929e-07)

        x = torch.FloatTensor(x)
        y = torch.LongTensor([y])

        gender = self.genders[idx]
        speaker_id = self.speaker_ids[idx]
        sample_id = self.sample_ids[idx]
        class_label = y_str

        return x, y, gender, speaker_id, sample_id, class_label


def get_data_loaders(args):
    logger.info('Using datasource_VAE_v1')

    path_train_emo = f"{args.path_emo}/train_dataset.json"
    path_test_emo = f"{args.path_emo}/test_dataset.json"
    dataset_train = Dataset(args, path_train_emo)
    dataset_test = Dataset(args, path_test_emo)

    """ Create WeightedRandomSampler for train dataloader """
    unique_classes = args.classes
    train_class_counts = np.zeros((len(unique_classes),), dtype=np.int32)
    train_class_labels = []
    for value in dataset_train.y_str:
        train_class_counts[unique_classes.index(value)] += 1
        train_class_labels.append(unique_classes.index(value))

    train_num_samples = sum(train_class_counts)
    train_class_weights = [train_num_samples / train_class_counts[i] for i in range(len(train_class_counts))]
    train_weights = [train_class_weights[train_class_labels[i]] for i in range(int(train_num_samples))]
    weighted_random_sampler = torch.utils.data.WeightedRandomSampler(
        weights=torch.DoubleTensor(train_weights),
        num_samples=int(train_num_samples)
    )

    data_loader_train = torch.utils.data.DataLoader(dataset_train,
                                                    batch_size=args.batch_size,
                                                    num_workers=args.data_workers,
                                                    drop_last=True if (train_num_samples % args.batch_size) != 0 else False,
                                                    pin_memory=True,
                                                    sampler=weighted_random_sampler)

    """ Balance test dataset to equal class label count """
    test_class_counts = np.zeros((len(args.classes),), dtype=np.int32)
    for value in dataset_train.y_str:
        test_class_counts[unique_classes.index(value)] += 1

    balanced_test_idxs = []
    balanced_test_counts = np.zeros((len(args.classes),), dtype=np.int32)
    min_class_count = np.min(test_class_counts)
    for idx, value in enumerate(dataset_test.y_str):
        label_idx = args.classes.index(value.lower())
        if balanced_test_counts[label_idx] < min_class_count:
            balanced_test_idxs.append(idx)
            balanced_test_counts[label_idx] += 1
    test_num_samples = sum(balanced_test_counts)
    balanced_test_dataset = torch.utils.data.Subset(dataset_test, indices=balanced_test_idxs)

    data_loader_test = torch.utils.data.DataLoader(balanced_test_dataset,
                                                   batch_size=args.test_batch_size,
                                                   num_workers=args.data_workers,
                                                   drop_last=True if (test_num_samples % args.test_batch_size) != 0 else False,
                                                   pin_memory=True)

    return data_loader_train, data_loader_test
