import torch
import torch.nn
import torch.utils.data
import torch.utils.data.sampler
import numpy as np

from loguru import logger
from modules.file_utils import FileUtils


class Dataset(torch.utils.data.dataset.Dataset):

    def __init__(self, args, datasource_json):
        super().__init__()
        self.args = args

        self.json_desc = FileUtils.readJSON(datasource_json)

        self.mmap_x = np.memmap(
            datasource_json.replace('.json', '.mmap'),
            dtype=np.float16,
            shape=tuple(self.json_desc['shape'])
        )

        self.lengths = self.json_desc['lengths']

        self.genders = self.json_desc['genders']
        self.sample_ids = self.json_desc['sample_ids']
        self.speaker_ids = self.json_desc['speaker_ids']
        self.data_samplerate = self.args.data_samplerate

        labels = self.json_desc['labels']
        self.y_str = [labels[i] for i in self.json_desc['class_idxes']]

    def __len__(self):
        return len(self.lengths)

    def __getitem__(self, idx):

        y_str = self.y_str[idx]
        if y_str in self.args.classes:
            y = self.args.classes.index(y_str)
        else:
            y = self.args.classes.index('other')

        x = self.mmap_x[idx]  # np.float16 (min eps = 0.000977, not enough for normalization)
        x = x.astype(np.float32)  # np.float32 (min eps = 1.1920929e-07)

        x = torch.FloatTensor(x)
        y = torch.LongTensor([y])

        if len(self.genders) != len(self.lengths):
            gender = self.genders[0]
        else:
            gender = self.genders[idx]

        speaker_id = self.speaker_ids[idx]
        sample_id = self.sample_ids[idx]

        if y_str in self.args.classes:
            class_label = y_str
        else:
            class_label = 'other'

        return x, y, gender, speaker_id, sample_id, class_label


def get_data_loaders(args):
    logger.info('Using datasource_contrastive_v1')

    dataset_train = Dataset(args, args.path_train_emo)
    dataset_test = Dataset(args, args.path_test_emo)

    targets = dataset_train.json_desc['class_idxes']
    counts = np.bincount(np.array(targets))
    sampler_weights = 1 / counts

    sample_weight = torch.FloatTensor([sampler_weights[int(t)] for t in targets])
    sampler = torch.utils.data.WeightedRandomSampler(sample_weight, len(sample_weight), replacement=True)

    logger.info('train dataset')
    data_loader_train = torch.utils.data.DataLoader(dataset_train,
                                                    batch_size=args.batch_size,
                                                    num_workers=args.data_workers,
                                                    pin_memory=True,
                                                    drop_last=True,
                                                    sampler=sampler)
    logger.info('test dataset')
    data_loader_test = torch.utils.data.DataLoader(dataset_test,
                                                   batch_size=args.batch_size,
                                                   num_workers=args.data_workers,
                                                   pin_memory=True,
                                                   drop_last=True)

    return data_loader_train, data_loader_test
