import torch
import torch.nn
import torch.utils.data
import torch.utils.data.sampler
import librosa
import numpy as np

from torch.utils.data.sampler import WeightedRandomSampler
from loguru import logger
from modules.file_utils import FileUtils


class Dataset(torch.utils.data.dataset.Dataset):

    def __init__(self, args, datasource_json):
        super().__init__()
        self.args = args

        self.json_desc = FileUtils.readJSON(datasource_json)

        self.mmap_x = np.memmap(
            datasource_json.replace('.json', '.mmap'),
            dtype=np.float16,
            shape=tuple(self.json_desc['shape'])
        )

        self.lengths = self.json_desc['lengths']

        self.genders = self.json_desc['genders']
        self.sample_ids = self.json_desc['sample_ids']
        self.speaker_ids = self.json_desc['speaker_ids']
        self.data_samplerate = self.args.data_samplerate
        self.class_idxes = self.json_desc['class_idxes']

        labels = self.json_desc['classes']
        self.y_str = [labels[i] for i in self.json_desc['class_idxes']]
        for label in set(self.y_str):
            logger.info(f'For class {label}: {self.y_str.count(label)} samples')

    def __len__(self):
        return len(self.lengths)

    def __getitem__(self, idx):

        y_str = self.y_str[idx]
        if y_str in self.args.classes:
            y = self.args.classes.index(y_str)
            class_label = y_str
        else:
            y = self.args.classes.index('other')
            class_label = 'other'

        x = self.mmap_x[idx]  # np.float16 (min eps = 0.000977, not enough for normalization)
        x = x.astype(np.float32)  # np.float32 (min eps = 1.1920929e-07)
        if self.args.downsample_input:
            x = librosa.resample(y=x, orig_sr=16_000, target_sr=8_000)
            x = librosa.resample(y=x, orig_sr=8_000, target_sr=16_000)

        x = torch.FloatTensor(x)
        y = torch.LongTensor([y])

        # gender = self.genders[idx]
        gender = 'any'
        speaker_id = self.speaker_ids[idx]
        sample_id = self.sample_ids[idx]

        return x, y, gender, speaker_id, sample_id, class_label


def get_data_loaders(args):
    logger.info('Using datasource_v1')

    dataset_train = Dataset(args, args.path_train_emo)
    dataset_test = Dataset(args, args.path_test_emo)

    y_train_indices = dataset_train.class_idxes
    class_sample_count = np.array(
        [len(np.where(y_train_indices == t)[0]) for t in np.unique(y_train_indices)])

    if args.method == 'division':
        weight = 1.0 / class_sample_count
    elif args.method == 'division_evalds':
        weight = (args.alpha / class_sample_count) * np.sum(class_sample_count)
    elif args.method == 'sum':
        weight = 1 - class_sample_count / np.sum(class_sample_count)
    elif args.method == 'max':
        weight = 1 + args.alpha - class_sample_count / np.max(class_sample_count)


    samples_weight = np.array([weight[t] for t in y_train_indices])
    samples_weight = torch.from_numpy(samples_weight)
    sampler = WeightedRandomSampler(samples_weight.type('torch.DoubleTensor'), len(samples_weight))

    logger.info('train dataset')
    data_loader_train = torch.utils.data.DataLoader(dataset_train,
                                                    batch_size=args.batch_size,
                                                    num_workers=args.data_workers,
                                                    pin_memory=True,
                                                    drop_last=False,
                                                    sampler=sampler)
    logger.info(f'{len(data_loader_train)}')
    logger.info('test dataset')
    data_loader_test = torch.utils.data.DataLoader(dataset_test,
                                                   batch_size=args.batch_size,
                                                   num_workers=args.data_workers,
                                                   pin_memory=True,
                                                   drop_last=False)
    logger.info(f'{len(data_loader_test)}')
    return data_loader_train, data_loader_test
