import math
import torch
from torch import nn
from torch.nn import functional as F

def sinc(t):
    """sinc.

    :param t: the input tensor
    """
    return torch.where(t == 0, torch.tensor(1., device=t.device, dtype=t.dtype), torch.sin(t) / t)


def kernel_upsample2(zeros=56):
    """kernel_upsample2.

    """
    win = torch.hann_window(4 * zeros + 1, periodic=False)
    winodd = win[1::2]
    t = torch.linspace(-zeros + 0.5, zeros - 0.5, 2 * zeros)
    t *= math.pi
    kernel = (sinc(t) * winodd).view(1, 1, -1)
    return kernel


def upsample2(x, zeros=56):
    """
    Upsampling the input by 2 using sinc interpolation.
    Smith, Julius, and Phil Gossett. "A flexible sampling-rate conversion method."
    ICASSP'84. IEEE International Conference on Acoustics, Speech, and Signal Processing.
    Vol. 9. IEEE, 1984.
    """
    *other, time = x.shape
    kernel = kernel_upsample2(zeros).to(x)
    out = F.conv1d(x.view(-1, 1, time), kernel, padding=zeros)[..., 1:].view(*other, time)
    y = torch.stack([x, out], dim=-1)
    return y.view(*other, -1)


def kernel_downsample2(zeros=56):
    """kernel_downsample2.

    """
    win = torch.hann_window(4 * zeros + 1, periodic=False)
    winodd = win[1::2]
    t = torch.linspace(-zeros + 0.5, zeros - 0.5, 2 * zeros)
    t.mul_(math.pi)
    kernel = (sinc(t) * winodd).view(1, 1, -1)
    return kernel


def downsample2(x, zeros=56):
    """
    Downsampling the input by 2 using sinc interpolation.
    Smith, Julius, and Phil Gossett. "A flexible sampling-rate conversion method."
    ICASSP'84. IEEE International Conference on Acoustics, Speech, and Signal Processing.
    Vol. 9. IEEE, 1984.
    """
    if x.shape[-1] % 2 != 0:
        x = F.pad(x, (0, 1))
    xeven = x[..., ::2]
    xodd = x[..., 1::2]
    *other, time = xodd.shape
    kernel = kernel_downsample2(zeros).to(x)
    out = xeven + F.conv1d(xodd.view(-1, 1, time), kernel, padding=zeros)[..., :-1].view(
        *other, time)
    return out.view(*other, -1).mul(0.5)


class BLSTM(nn.Module):
    def __init__(self, dim, layers=2, bi=True):
        super().__init__()
        self.lstm = nn.LSTM(bidirectional=bi, num_layers=layers, hidden_size=dim, input_size=dim)
        self.linear = None
        if bi:
            self.linear = nn.Linear(2 * dim, dim)

    def forward(self, x, hidden=None):
        x, hidden = self.lstm(x, hidden)
        if self.linear:
            x = self.linear(x)
        return x, hidden


def rescale_conv(conv, reference):
    std = conv.weight.std().detach()
    scale = (std / reference)**0.5
    conv.weight.data /= scale
    if conv.bias is not None:
        conv.bias.data /= scale


def rescale_module(module, reference):
    for sub in module.modules():
        if isinstance(sub, (nn.Conv1d, nn.ConvTranspose1d)):
            rescale_conv(sub, reference)


class Model(nn.Module):
    """
    Demucs speech enhancement model.
    Args:
        - chin (int): number of input channels.
        - chout (int): number of output channels.
        - hidden (int): number of initial hidden channels.
        - depth (int): number of layers.
        - kernel_size (int): kernel size for each layer.
        - stride (int): stride for each layer.
        - causal (bool): if false, uses BiLSTM instead of LSTM.
        - resample (int): amount of resampling to apply to the input/output.
            Can be one of 1, 2 or 4.
        - growth (float): number of channels is multiplied by this for every layer.
        - max_hidden (int): maximum number of channels. Can be useful to
            control the size/speed of the model.
        - normalize (bool): if true, normalize the input.
        - glu (bool): if true uses GLU instead of ReLU in 1x1 convolutions.
        - rescale (float): controls custom weight initialization.
            See https://arxiv.org/abs/1911.13254.
        - floor (float): stability flooring when normalizing.

    """
    def __init__(self, args):

        super().__init__()
        self.args = args

        if args.resample not in [0.25, 0.5, 1, 2, 4]:
            raise ValueError("Resample should be 1, 2 or 4.")

        self.chin = 1
        self.depth = self.args.depth
        self.hidden = self.args.hidden_size
        self.kernel_size = self.args.kernel_size
        self.stride = self.args.stride
        self.causal = self.args.causal
        self.floor = self.args.floor
        self.resample = self.args.resample
        self.normalize = self.args.normalize
        self.growth = self.args.growth
        self.rescale = self.args.rescale
        self.max_hidden = self.args.max_hidden
        self.embedding_size = self.args.embedding_size
        self.dropout = self.args.dropout_rate

        self.data_len = self.args.input_size * self.resample

        self.encoder = nn.ModuleList()
        activation = nn.GLU(1) if self.args.is_glu else nn.ReLU()
        ch_scale = 2 if self.args.is_glu else 1

        chin = self.chin
        hidden = self.hidden
        self.final_hidden = self.hidden
        for index in range(self.depth):
            encode = []
            encode += [
                nn.Conv1d(chin, hidden, self.kernel_size, self.stride),
                nn.ReLU(),
                nn.Conv1d(hidden, hidden * ch_scale, 1), activation,
                nn.BatchNorm1d(num_features=hidden)
            ]
            self.encoder.append(nn.Sequential(*encode))
            self.data_len = int((self.data_len - self.kernel_size) / self.stride + 1)
            self.final_hidden = hidden

            chin = hidden
            hidden = min(int(self.growth * hidden), self.max_hidden)

        self.lstm = BLSTM(chin, bi=not self.causal)
        self.mlp = torch.nn.Sequential(
            torch.nn.Dropout(self.dropout),
            torch.nn.Linear(in_features=self.final_hidden*self.data_len, out_features=self.final_hidden),
            torch.nn.LeakyReLU(),
            torch.nn.Dropout(self.dropout),
            torch.nn.Linear(in_features=self.final_hidden, out_features=self.embedding_size)
        )

        if self.rescale:
            rescale_module(self, reference=self.rescale)

    def forward(self, mix):
        if mix.dim() == 2:
            mix = mix.unsqueeze(1)

        if self.normalize:
            mono = mix.mean(dim=1, keepdim=True)
            std = mono.std(dim=-1, keepdim=True)
            mix = mix / (self.floor + std)
        x = mix
        if self.resample == 2:
            x = upsample2(x)
        elif self.resample == 4:
            x = upsample2(x)
            x = upsample2(x)
        elif self.resample == 0.5:
            x = downsample2(x)
        elif self.resample == 0.25:
            x = downsample2(x)
            x = downsample2(x)

        for encode in self.encoder:
            x = encode.forward(x)
        x = x.permute(2, 0, 1)
        x, _ = self.lstm.forward(x)
        x = x.permute(1, 2, 0).contiguous()

        # # selecting only max over seq dimension
        # # other possible ways to make model easily streamable?
        # x = torch.max(x, dim=2).values
        x = x.view(-1, self.final_hidden * self.data_len)
        x = self.mlp.forward(x)
        norm = torch.norm(x.detach(), p=2, dim=-1, keepdim=True)
        x = x / norm

        return x
