import torch
import torch.nn as nn
from torch.nn import functional as F
import torch.utils.data

from torch_audiomentations import Compose, LowPassFilter, HighPassFilter, AddColoredNoise, Gain, Shift
from torchaudio.transforms import MFCC, Spectrogram, MelSpectrogram, AmplitudeToDB


class Encoder(nn.Module):
    def __init__(self, hidden_size):
        super().__init__()
        self.cnn = nn.Sequential(
            nn.Conv2d(in_channels=2, out_channels=8, kernel_size=3, stride=1, padding=1),
            nn.BatchNorm2d(num_features=8),
            nn.ReLU(),
            nn.MaxPool2d(kernel_size=4, stride=2, padding=1, dilation=1),
            nn.Conv2d(in_channels=8, out_channels=16, kernel_size=3, stride=1, padding=1),
            nn.BatchNorm2d(num_features=16),
            nn.ReLU(),
            nn.MaxPool2d(kernel_size=4, stride=2, padding=1, dilation=1),
            nn.Conv2d(in_channels=16, out_channels=32, kernel_size=3, stride=1, padding=1),
            nn.BatchNorm2d(num_features=32),
            nn.ReLU(),
            nn.MaxPool2d(kernel_size=4, stride=2, padding=1, dilation=1),
            nn.Conv2d(in_channels=32, out_channels=64, kernel_size=3, stride=1, padding=1),
            nn.BatchNorm2d(num_features=64),
            nn.ReLU(),
            nn.MaxPool2d(kernel_size=4, stride=2, padding=1, dilation=1),
            nn.Conv2d(in_channels=64, out_channels=64, kernel_size=3, stride=1, padding=1)
        )
        self.lienar_mu = nn.Linear(in_features=8*8*64, out_features=hidden_size)
        self.linear_sigma = nn.Linear(in_features=8*8*64, out_features=hidden_size)

    def forward(self, x):
        x = self.cnn.forward(x)
        x = torch.flatten(x, start_dim=1)
        mu = self.lienar_mu.forward(x)
        sigma = self.linear_sigma.forward(x)
        return mu, sigma


class Decoder(nn.Module):
    def __init__(self, c_dim, hidden_size):
        super().__init__()
        self.decoder_lin = nn.Sequential(
            nn.Linear(in_features=hidden_size + c_dim, out_features=512),
            nn.BatchNorm1d(num_features=512),
            nn.ReLU(),
            nn.Dropout(p=0.1),
            nn.Linear(in_features=512, out_features=8 * 8 * 64),
            nn.ReLU()
        )
        self.unflatten = nn.Unflatten(dim=1, unflattened_size=(64, 8, 8))
        self.decoder_conv = nn.Sequential(
            nn.Conv2d(in_channels=64, out_channels=64, kernel_size=3, stride=1, padding=1),
            nn.BatchNorm2d(num_features=64),
            nn.ReLU(),
            nn.Upsample(scale_factor=2),
            nn.Conv2d(in_channels=64, out_channels=32, kernel_size=3, stride=1, padding=1),
            nn.BatchNorm2d(num_features=32),
            nn.ReLU(),
            nn.Upsample(scale_factor=2),
            nn.Conv2d(in_channels=32, out_channels=16, kernel_size=3, stride=1, padding=1),
            nn.BatchNorm2d(num_features=16),
            nn.ReLU(),
            nn.Upsample(scale_factor=2),
            nn.Conv2d(in_channels=16, out_channels=8, kernel_size=3, stride=1, padding=1),
            nn.BatchNorm2d(num_features=8),
            nn.ReLU(),
            nn.Upsample(scale_factor=2),
            nn.Conv2d(in_channels=8, out_channels=1, kernel_size=3, stride=1, padding=1),
            nn.Sigmoid()
        )

    def forward(self, x):
        x = self.decoder_lin(x)
        x = self.unflatten(x)
        x = self.decoder_conv(x)
        return x


class Model(nn.Module):
    def __init__(self, args):
        super().__init__()
        self.args = args
        self.n_classes = args.n_classes
        self.img_size = args.img_size
        self.hidden_size = args.hidden_size
        self.encoder = Encoder(hidden_size=self.hidden_size)
        self.decoder = Decoder(c_dim=self.n_classes, hidden_size=self.hidden_size)

        self.embed_class = nn.Linear(in_features=self.n_classes, out_features=self.img_size * self.img_size)

        self.augmentations = Compose(
            transforms=[
                Gain(
                    min_gain_in_db=-15.0,
                    max_gain_in_db=5.0,
                    p=self.args.each_augment_proba,
                ),
                AddColoredNoise(p=self.args.each_augment_proba),
                Shift(min_shift=-0.25, max_shift=0.25, p=self.args.each_augment_proba),
                LowPassFilter(p=self.args.each_augment_proba),
                HighPassFilter(p=self.args.each_augment_proba)
            ]
        )

        self.downsample = torch.nn.Upsample(size=(self.img_size, self.img_size))
        if self.args.spectrogram_type == "spectrogram":
            self.transform = Spectrogram(n_fft=self.args.n_fft,
                                         win_length=self.args.win_length,
                                         hop_length=self.args.hop_length).to(self.args.device)
        elif self.args.spectrogram_type == "melspectrogram":
            self.transform = MelSpectrogram(sample_rate=self.args.data_samplerate,
                                            n_fft=self.args.n_fft,
                                            n_mels=self.args.n_mels,
                                            win_length=self.args.win_length,
                                            hop_length=self.args.hop_length).to(self.args.device)
        elif self.args.spectrogram_type == "mfcc":
            self.transform = MFCC(sample_rate=self.args.data_samplerate,
                                  n_mfcc=self.args.n_mfcc,
                                  melkwargs={
                                      "n_fft": self.args.n_fft,
                                      "n_mels": self.args.n_mels,
                                      "hop_length": self.args.hop_length,
                                      "win_length": self.args.win_length
                                  }).to(self.args.device)
        self.amplitude_to_db = AmplitudeToDB()

    def standardize(self, batch, scope='sample'):
        eps = 1e-8
        if scope == 'sample':
            for idx, sample in enumerate(batch):
                mean = torch.mean(sample)
                std = torch.std(sample)
                batch[idx] = (batch[idx] - mean) / (std + eps)
        elif scope == 'batch':
            mean = torch.mean(batch)
            std = torch.std(batch)
            batch = (batch - mean) / (std + eps)
        return batch

    def idx2onehot(self, idx, n):
        if idx.dim() == 1:
            idx = idx.unsqueeze(1)
        onehot = torch.zeros(idx.size(0), n, device=self.args.device)
        onehot.scatter_(1, idx, 1)
        return onehot

    def normalize(self, x):
        min = x.min()
        max = x.max()
        x = (x - min) / (max - min)
        return x

    def reparameterize(self, mu, logvar):
        std = torch.exp(0.5 * logvar)
        eps = torch.randn_like(std)
        return eps * std + mu

    def sample(self, num_samples, labels):
        z = torch.randn(num_samples, self.hidden_size)
        z = z.to(self.args.device)

        labels_one_hot = self.idx2onehot(labels, n=self.n_classes)
        z = torch.cat([z, labels_one_hot], dim=1)
        samples = self.decoder(z)
        return samples

    def forward(self, x, labels):
        x = x.to(self.args.device).unsqueeze(dim=1)
        labels = labels.to(self.args.device)

        if self.training:
            x = self.augmentations(x, sample_rate=self.args.data_samplerate)

        x = self.standardize(x, scope='sample')

        x_in = self.transform(x)
        x_in = self.amplitude_to_db(x_in)

        x_in = self.downsample(x_in)
        x_in = self.normalize(x_in)

        labels_one_hot = self.idx2onehot(labels, n=self.n_classes)
        embedded_class = self.embed_class.forward(labels_one_hot)
        embedded_class = embedded_class.view(-1, self.img_size, self.img_size).unsqueeze(dim=1)
        x_labels_emb = torch.cat([x_in, embedded_class], dim=1)

        mu, log_var = self.encoder(x_labels_emb)

        z = self.reparameterize(mu, log_var)

        z = torch.cat([z, labels_one_hot], dim=1)
        z_dec = self.decoder(z)

        return z_dec, x_in, mu, log_var

